#include "text.h"

Text::Text(CGapiDisplay* pDisplay)
{
	m_pDisplay	= pDisplay;

	::ZeroMemory(&m_region, sizeof(SIZE));

	m_padding		= TEXT_DEFAULT_PADDING;
	m_lineSpacing	= TEXT_DEFAULT_SPACING;

	m_alignment		= NULL;
	::ZeroMemory(&m_alignPos, sizeof(POINT));

	m_pFont			= NULL;
}

Text::~Text()
{
	// No need to do anything.
}

void Text::Draw(CGapiSurface* pBuffer, LONG x, LONG y) const
{
	if (m_text.empty()) {
		return;	// No reason to draw an empty string.
	}

	CGapiBitmapFont* pFont = this->GetDrawingFont();
	DWORD fontHeight = pFont->GetHeight();

	// Adjust for padding.
	LONG xPos = x + m_alignPos.x;
	LONG yPos = y + m_alignPos.y;

	// Draw text line by line. It should already be formatted,
	// so drawing should be as efficient as possible.

	tstring_t strTok(m_text);
	TCHAR* pLineTok			= NULL;
	const TCHAR pTokens[]	= { _T("\n")[0], TEXT_LINE_SEPERATOR[0], NULL };

	pLineTok = _tcstok(const_cast<TCHAR*>(strTok.c_str()), pTokens);	// First line.

	while (pLineTok != NULL && (yPos + (LONG) fontHeight) < (y + m_region.cy - (LONG) m_padding))
	{
		// Draw line.
		if (pBuffer->DrawText(xPos, yPos, pLineTok, pFont, m_alignment, NULL, NULL, NULL) != S_OK) {
			return;
		};

		// Move to next line.
		pLineTok = _tcstok(NULL, pTokens);
		yPos += fontHeight + m_lineSpacing;
	}
}

void Text::SetText(const tstring_t& text)
{
	m_text = text;
}

const tstring_t& Text::GetText() const
{
	return m_text;
}

void Text::SetRegion(DWORD width, DWORD height)
{
	m_region.cx = width;
	m_region.cy = height;
}

const SIZE& Text::GetRegion() const
{
	return m_region;
}

void Text::SetPadding(DWORD padding)
{
	m_padding = padding;
}

DWORD Text::GetPadding() const
{
	return m_padding;
}

void Text::SetLineSpacing(DWORD spacing)
{
	m_lineSpacing = spacing;
}

DWORD Text::GetLineSpacing() const
{
	return m_lineSpacing;
}

void Text::SetAlignment(DWORD alignment)
{
	m_alignment = alignment;
}

DWORD Text::GetAlignment() const
{
	return m_alignment;
}

void Text::SetFont(CGapiBitmapFont* pFont)
{
	m_pFont = pFont;
}

void Text::Format()
{
	// Assume the text has been formatted before,
	// thus all line seperators need to be removed.
	this->RemoveLineSeperators();

	CGapiBitmapFont* pFont		= this->GetDrawingFont();
	tstring_t::size_type index	= 0;
	DWORD charSpacing			= 0;
	DWORD wordWidth				= 0;
	DWORD lineWidth				= 0;
	DWORD numOfLines			= 0;
	DWORD padding				= m_padding << 1;

	DWORD spaceWidth;
	pFont->GetCharWidth(_T(" ")[0], &spaceWidth);

	tstring_t strWord;
	strWord.reserve(TEXT_WORD_LENGTH);

	tstring_t::size_type lineBeginPos = 0, lineEndPos, wordBeginPos = 0, wordEndPos;
	lineEndPos = m_text.find(_T("\n"));	// First line.

	// Go through all the lines in the text.
	while (true)
	{
		// Move to new line.
		numOfLines++;
		lineWidth	= padding;

		// First word in line.
		wordEndPos	= m_text.find(_T(" "), lineBeginPos);

		// Go through all the words in the line.
		while (true)
		{
			// Get size of the entire word.
			strWord = m_text.substr(wordBeginPos, wordEndPos - wordBeginPos);
			pFont->GetStringWidth(strWord.c_str(), &wordWidth);

			// Check for word spacing.
			if (lineWidth > padding) {
				lineWidth += spaceWidth;
			}

			// Check word wrap.
			lineWidth += wordWidth;
			if (lineWidth > (DWORD) m_region.cx) {
				if (wordBeginPos != 0) {
					// Mark line seperation.
					m_text[wordBeginPos - 1] = TEXT_LINE_SEPERATOR[0];
				}

				// Move to new line and add the word to it.
				numOfLines++;
				lineWidth = padding + wordWidth;
			}

			if (wordEndPos == tstring_t::npos || wordEndPos > lineEndPos) {
				break;	// No more words on this line.
			}

			// Move to next word.
			wordBeginPos	= wordEndPos + 1;
			wordEndPos		= m_text.find(_T(" "), wordBeginPos);
		}

		if (lineEndPos == tstring_t::npos) {
			break;	// No more lines to be found.
		}

		// Move to next line.
		lineBeginPos	= lineEndPos + 1;
		lineEndPos		= m_text.find(_T("\n"), lineBeginPos);
	}

	// Properly align text to the text region.
	this->AlignText(numOfLines, pFont->GetHeight());
}

void Text::RemoveLineSeperators()
{
	tstring_t::size_type pos;
	pos = m_text.find(TEXT_LINE_SEPERATOR);	// First line seperator.

	// Go until no line seperators are found.
	while (pos != tstring_t::npos)
	{
		m_text[pos] = _T(" ")[0];	// Insert space where seperator is at.

		// Get next seperator.
		pos = m_text.find(TEXT_LINE_SEPERATOR, pos + 1);
	}
}

void Text::AlignText(DWORD numOfLines, DWORD fontHeight)
{
	switch(m_alignment)
	{
		case GDDRAWTEXT_LEFT:
		{
			m_alignPos.x = m_padding;
			m_alignPos.y = m_padding;
			break;
		}
		case GDDRAWTEXT_CENTER:
		{
			m_alignPos.x = m_region.cx >> 1;

			DWORD textHeight = numOfLines * (fontHeight + m_lineSpacing) - m_lineSpacing;
			if (textHeight > m_region.cy - (m_padding << 1)) {
				// Text won't event fit, so don't align y axis.
				m_alignPos.y = m_padding;
			}
			else {
				m_alignPos.y = (m_region.cy >> 1) - (textHeight >> 1);
			}

			break;
		}
		case GDDRAWTEXT_RIGHT:
		{
			m_alignPos.x = m_region.cx - m_padding;
			m_alignPos.y = m_padding;
			break;
		}
		default:	// Should never happen.
			break;
	}
}

CGapiBitmapFont* Text::GetDrawingFont() const
{
	if (m_pFont == NULL) {
		return m_pDisplay->GetSystemFontBorder();
	}
	else {
		return m_pFont;
	}
}

			//wordWidth = 0;
			//for (index = wordBeginPos; index < wordEndPos && index < lineEndPos && index < m_text.size(); ++index) {
			//	pFont->GetCharWidth(m_text[index], &charWidth);
			//	wordWidth += charWidth;

			//	// Adjust width based on spacing between chars.
			//	if (index != 0) {
			//		pFont->GetSpacing(m_text[index-1], m_text[index], &charSpacing);
			//		wordWidth += charSpacing;
			//	}
			//}
